@@ -1,7 +1,11 @@
 module RBS
   module Prototype
     class RB
-      Context = Struct.new(:module_function, :singleton, :namespace, keyword_init: true) do
+      include Helpers
+
+      Context = _ = Struct.new(:module_function, :singleton, :namespace, keyword_init: true) do
+        # @implements Context
+
         def self.initial(namespace: Namespace.root)
           self.new(module_function: false, singleton: false, namespace: namespace)
         end
@@ -33,9 +37,12 @@ module RBS
       end
 
       def decls
+        # @type var decls: Array[AST::Declarations::t]
         decls = []
 
-        top_decls, top_members = source_decls.partition {|decl| decl.is_a?(AST::Declarations::Base) }
+        # @type var top_decls: Array[AST::Declarations::t]
+        # @type var top_members: Array[AST::Members::t]
+        top_decls, top_members = _ = source_decls.partition {|decl| decl.is_a?(AST::Declarations::Base) }
 
         decls.push(*top_decls)
 
@@ -47,7 +54,7 @@ module RBS
             annotations: [],
             comment: nil,
             location: nil,
-            type_params: AST::Declarations::ModuleTypeParams.empty
+            type_params: []
           )
           decls << top
         end
@@ -56,11 +63,18 @@ module RBS
       end
 
       def parse(string)
+        # @type var comments: Hash[Integer, AST::Comment]
         comments = Ripper.lex(string).yield_self do |tokens|
+          code_lines = {}
           tokens.each.with_object({}) do |token, hash|
-            if token[1] == :on_comment
+            case token[1]
+            when :on_sp, :on_ignored_nl
+              # skip
+            when :on_comment
               line = token[0][0]
-              body = token[2][2..-1]
+              # skip like `module Foo # :nodoc:`
+              next if code_lines[line]
+              body = token[2][2..-1] or raise
 
               body = "\n" if body.empty?
 
@@ -72,6 +86,8 @@ module RBS
               else
                 hash[line] = comment
               end
+            else
+              code_lines[token[0][0]] = true
             end
           end
         end
@@ -82,11 +98,19 @@ module RBS
       def process(node, decls:, comments:, context:)
         case node.type
         when :CLASS
-          class_name, super_class, *class_body = node.children
+          class_name, super_class_node, *class_body = node.children
+          super_class_name = const_to_name(super_class_node, context: context)
+          super_class =
+            if super_class_name
+              AST::Declarations::Class::Super.new(name: super_class_name, args: [], location: nil)
+            else
+              # Give up detect super class e.g. `class Foo < Struct.new(:bar)`
+              nil
+            end
           kls = AST::Declarations::Class.new(
-            name: const_to_name(class_name),
-            super_class: super_class && AST::Declarations::Class::Super.new(name: const_to_name(super_class), args: []),
-            type_params: AST::Declarations::ModuleTypeParams.empty,
+            name: const_to_name!(class_name),
+            super_class: super_class,
+            type_params: [],
             members: [],
             annotations: [],
             location: nil,
@@ -105,8 +129,8 @@ module RBS
           module_name, *module_body = node.children
 
           mod = AST::Declarations::Module.new(
-            name: const_to_name(module_name),
-            type_params: AST::Declarations::ModuleTypeParams.empty,
+            name: const_to_name!(module_name),
+            type_params: [],
             self_types: [],
             members: [],
             annotations: [],
@@ -129,38 +153,44 @@ module RBS
             RBS.logger.warn "`class <<` syntax with not-self may be compiled to incorrect code: #{this}"
           end
 
+          accessibility = current_accessibility(decls)
+
           ctx = Context.initial.tap { |ctx| ctx.singleton = true }
           process_children(body, decls: decls, comments: comments, context: ctx)
 
+          decls << accessibility
+
         when :DEFN, :DEFS
-            if node.type == :DEFN
-              def_name, def_body = node.children
-              kind = context.method_kind
-            else
-              _, def_name, def_body = node.children
-              kind = :singleton
-            end
+          # @type var kind: Context::method_kind
 
-            types = [
-              MethodType.new(
-                type_params: [],
-                type: function_type_from_body(def_body),
-                block: block_from_body(def_body),
-                location: nil
-              )
-            ]
+          if node.type == :DEFN
+            def_name, def_body = node.children
+            kind = context.method_kind
+          else
+            _, def_name, def_body = node.children
+            kind = :singleton
+          end
 
-            member = AST::Members::MethodDefinition.new(
-              name: def_name,
-              location: nil,
-              annotations: [],
-              types: types,
-              kind: kind,
-              comment: comments[node.first_lineno - 1],
-              overload: false
+          types = [
+            MethodType.new(
+              type_params: [],
+              type: function_type_from_body(def_body, def_name),
+              block: block_from_body(def_body),
+              location: nil
             )
+          ]
 
-            decls.push member unless decls.include?(member)
+          member = AST::Members::MethodDefinition.new(
+            name: def_name,
+            location: nil,
+            annotations: [],
+            types: types,
+            kind: kind,
+            comment: comments[node.first_lineno - 1],
+            overload: false
+          )
+
+          decls.push member unless decls.include?(member)
 
         when :ALIAS
           new_name, old_name = node.children.map { |c| literal_to_symbol(c) }
@@ -181,7 +211,7 @@ module RBS
           case node.children[0]
           when :include
             args.each do |arg|
-              if (name = const_to_name(arg))
+              if (name = const_to_name(arg, context: context))
                 decls << AST::Members::Include.new(
                   name: name,
                   args: [],
@@ -191,6 +221,18 @@ module RBS
                 )
               end
             end
+          when :prepend
+            args.each do |arg|
+              if (name = const_to_name(arg, context: context))
+                decls << AST::Members::Prepend.new(
+                  name: name,
+                  args: [],
+                  annotations: [],
+                  location: nil,
+                  comment: comments[node.first_lineno - 1]
+                )
+              end
+            end
           when :extend
             args.each do |arg|
               if (name = const_to_name(arg, context: context))
@@ -263,8 +305,10 @@ module RBS
               module_func_context = context.dup.tap { |ctx| ctx.module_function = true }
               args.each do |arg|
                 if arg && (name = literal_to_symbol(arg))
-                  if i = find_def_index_by_name(decls, name)
-                    decls[i] = decls[i].update(kind: :singleton_instance)
+                  if (i, defn = find_def_index_by_name(decls, name))
+                    if defn.is_a?(AST::Members::MethodDefinition)
+                      decls[i] = defn.update(kind: :singleton_instance)
+                    end
                   end
                 elsif arg
                   process arg, decls: decls, comments: comments, context: module_func_context
@@ -278,7 +322,7 @@ module RBS
             else
               args.each do |arg|
                 if arg && (name = literal_to_symbol(arg))
-                  if i = find_def_index_by_name(decls, name)
+                  if (i, _ = find_def_index_by_name(decls, name))
                     current = current_accessibility(decls, i)
                     if current != accessibility
                       decls.insert(i + 1, current)
@@ -312,12 +356,19 @@ module RBS
                        when node.children[0].is_a?(Symbol)
                          TypeName.new(name: node.children[0], namespace: Namespace.empty)
                        else
-                         const_to_name(node.children[0])
+                         const_to_name!(node.children[0])
                        end
 
+          value_node = node.children.last
+          type = if value_node.nil?
+                  # Give up type prediction when node is MASGN.
+                  Types::Bases::Any.new(location: nil)
+                else
+                  literal_to_type(value_node)
+                end
           decls << AST::Declarations::Constant.new(
             name: const_name,
-            type: node_type(node.children.last),
+            type: type,
             location: nil,
             comment: comments[node.first_lineno - 1]
           )
@@ -333,13 +384,13 @@ module RBS
         end
       end
 
-      def const_to_name(node, context: nil)
-        case node&.type
+      def const_to_name!(node)
+        case node.type
         when :CONST
           TypeName.new(name: node.children[0], namespace: Namespace.empty)
         when :COLON2
           if node.children[0]
-            namespace = const_to_name(node.children[0]).to_namespace
+            namespace = const_to_name!(node.children[0]).to_namespace
           else
             namespace = Namespace.empty
           end
@@ -347,8 +398,19 @@ module RBS
           TypeName.new(name: node.children[1], namespace: namespace)
         when :COLON3
           TypeName.new(name: node.children[0], namespace: Namespace.root)
-        when :SELF
-          context&.then { |c| c.namespace.to_type_name }
+        else
+          raise
+        end
+      end
+
+      def const_to_name(node, context:)
+        if node
+          case node.type
+          when :SELF
+            context.namespace.to_type_name
+          when :CONST, :COLON2, :COLON3
+            const_to_name!(node)
+          end
         end
       end
 
@@ -361,24 +423,16 @@ module RBS
         end
       end
 
-      def each_node(nodes)
-        nodes.each do |child|
-          if child.is_a?(RubyVM::AbstractSyntaxTree::Node)
-            yield child
-          end
-        end
-      end
-
-      def each_child(node, &block)
-        each_node node.children, &block
-      end
-
-      def function_type_from_body(node)
+      def function_type_from_body(node, def_name)
         table_node, args_node, *_ = node.children
 
-        pre_num, _pre_init, opt, _first_post, post_num, _post_init, rest, kw, kwrest, _block = args_node.children
+        pre_num, _pre_init, opt, _first_post, post_num, _post_init, rest, kw, kwrest, _block = args_from_node(args_node)
 
-        return_type = function_return_type_from_body(node)
+        return_type = if def_name == :initialize
+                        Types::Bases::Void.new(location: nil)
+                      else
+                        function_return_type_from_body(node)
+                      end
 
         fun = Types::Function.empty(return_type)
 
@@ -391,12 +445,13 @@ module RBS
           name = lvasgn.children[0]
           fun.optional_positionals << Types::Function::Param.new(
             name: name,
-            type: node_type(lvasgn.children[1])
+            type: param_type(lvasgn.children[1])
           )
         end
 
         if rest
-          fun = fun.update(rest_positionals: Types::Function::Param.new(name: rest, type: untyped))
+          rest_name = rest == :* ? nil : rest # # For `def f(...) end` syntax
+          fun = fun.update(rest_positionals: Types::Function::Param.new(name: rest_name, type: untyped))
         end
 
         table_node.drop(fun.required_positionals.size + fun.optional_positionals.size + (fun.rest_positionals ? 1 : 0)).take(post_num).each do |name|
@@ -409,9 +464,9 @@ module RBS
 
           case value
           when nil, :NODE_SPECIAL_REQUIRED_KEYWORD
-            fun.required_keywords[name] = Types::Function::Param.new(name: name, type: untyped)
+            fun.required_keywords[name] = Types::Function::Param.new(name: nil, type: untyped)
           when RubyVM::AbstractSyntaxTree::Node
-            fun.optional_keywords[name] = Types::Function::Param.new(name: name, type: node_type(value))
+            fun.optional_keywords[name] = Types::Function::Param.new(name: nil, type: param_type(value))
           else
             raise "Unexpected keyword arg value: #{value}"
           end
@@ -426,28 +481,48 @@ module RBS
 
       def function_return_type_from_body(node)
         body = node.children[2]
-        return Types::Bases::Nil.new(location: nil) unless body
+        body_type(body)
+      end
 
-        if body.type == :BLOCK
-          return_stmts = any_node?(body) do |n|
-            n.type == :RETURN
-          end&.map do |return_node|
-            returned_value = return_node.children[0]
-            returned_value ? literal_to_type(returned_value) : Types::Bases::Nil.new(location: nil)
-          end || []
-          last_node = body.children.last
-          last_evaluated =  last_node ? literal_to_type(last_node) : Types::Bases::Nil.new(location: nil)
-          types_to_union_type([*return_stmts, last_evaluated])
+      def body_type(node)
+        return Types::Bases::Nil.new(location: nil) unless node
+
+        case node.type
+        when :IF, :UNLESS
+          if_unless_type(node)
+        when :BLOCK
+          block_type(node)
         else
-          literal_to_type(body)
+          literal_to_type(node)
         end
       end
 
+      def if_unless_type(node)
+        raise unless node.type == :IF || node.type == :UNLESS
+
+        _exp_node, true_node, false_node = node.children
+        types_to_union_type([body_type(true_node), body_type(false_node)])
+      end
+
+      def block_type(node)
+        raise unless node.type == :BLOCK
+
+        return_stmts = any_node?(node) do |n|
+          n.type == :RETURN
+        end&.map do |return_node|
+          returned_value = return_node.children[0]
+          returned_value ? literal_to_type(returned_value) : Types::Bases::Nil.new(location: nil)
+        end || []
+        last_node = node.children.last
+        last_evaluated =  last_node ? literal_to_type(last_node) : Types::Bases::Nil.new(location: nil)
+        types_to_union_type([*return_stmts, last_evaluated])
+      end
+
       def literal_to_type(node)
         case node.type
         when :STR
           lit = node.children[0]
-          if lit.match?(/\A[ -~]+\z/)
+          if lit.ascii_only?
             Types::Literal.new(literal: lit, location: nil)
           else
             BuiltinNames::String.instance_type
@@ -459,16 +534,16 @@ module RBS
         when :DREGX
           BuiltinNames::Regexp.instance_type
         when :TRUE
-          BuiltinNames::TrueClass.instance_type
+          Types::Literal.new(literal: true, location: nil)
         when :FALSE
-          BuiltinNames::FalseClass.instance_type
+          Types::Literal.new(literal: false, location: nil)
         when :NIL
           Types::Bases::Nil.new(location: nil)
         when :LIT
           lit = node.children[0]
           case lit
           when Symbol
-            if lit.match?(/\A[ -~]+\z/)
+            if lit.to_s.ascii_only?
               Types::Literal.new(literal: lit, location: nil)
             else
               BuiltinNames::Symbol.instance_type
@@ -480,15 +555,15 @@ module RBS
             Types::ClassInstance.new(name: type_name, args: [], location: nil)
           end
         when :ZLIST, :ZARRAY
-          BuiltinNames::Array.instance_type([untyped])
+          BuiltinNames::Array.instance_type(untyped)
         when :LIST, :ARRAY
           elem_types = node.children.compact.map { |e| literal_to_type(e) }
           t = types_to_union_type(elem_types)
-          BuiltinNames::Array.instance_type([t])
+          BuiltinNames::Array.instance_type(t)
         when :DOT2, :DOT3
           types = node.children.map { |c| literal_to_type(c) }
           type = range_element_type(types)
-          BuiltinNames::Range.instance_type([type])
+          BuiltinNames::Range.instance_type(type)
         when :HASH
           list = node.children[0]
           if list
@@ -516,7 +591,17 @@ module RBS
           else
             key_type = types_to_union_type(key_types)
             value_type = types_to_union_type(value_types)
-            BuiltinNames::Hash.instance_type([key_type, value_type])
+            BuiltinNames::Hash.instance_type(key_type, value_type)
+          end
+        when :SELF
+          Types::Bases::Self.new(location: nil)
+        when :CALL
+          receiver, method_name, * = node.children
+          case method_name
+          when :freeze, :tap, :itself, :dup, :clone, :taint, :untaint, :extend
+            literal_to_type(receiver)
+          else
+            untyped
           end
         else
           untyped
@@ -527,7 +612,9 @@ module RBS
         return untyped if types.empty?
 
         uniq = types.uniq
-        return uniq.first if uniq.size == 1
+        if uniq.size == 1
+          return uniq.first || raise
+        end
 
         Types::Union.new(types: uniq, location: nil)
       end
@@ -546,93 +633,13 @@ module RBS
         end.uniq
 
         if types.size == 1
-          types.first
+          types.first or raise
         else
           untyped
         end
       end
 
-      def block_from_body(node)
-        _, args_node, body_node = node.children
-
-        _pre_num, _pre_init, _opt, _first_post, _post_num, _post_init, _rest, _kw, _kwrest, block = args_node.children
-
-        method_block = nil
-
-        if block
-          method_block = Types::Block.new(
-            required: true,
-            type: Types::Function.empty(untyped)
-          )
-        end
-
-        if body_node
-          if (yields = any_node?(body_node) {|n| n.type == :YIELD })
-            method_block = Types::Block.new(
-              required: true,
-              type: Types::Function.empty(untyped)
-            )
-
-            yields.each do |yield_node|
-              array_content = yield_node.children[0]&.children&.compact || []
-
-              positionals, keywords = if keyword_hash?(array_content.last)
-                                        [array_content.take(array_content.size - 1), array_content.last]
-                                      else
-                                        [array_content, nil]
-                                      end
-
-              if (diff = positionals.size - method_block.type.required_positionals.size) > 0
-                diff.times do
-                  method_block.type.required_positionals << Types::Function::Param.new(
-                    type: untyped,
-                    name: nil
-                  )
-                end
-              end
-
-              if keywords
-                keywords.children[0].children.each_slice(2) do |key_node, value_node|
-                  if key_node
-                    key = key_node.children[0]
-                    method_block.type.required_keywords[key] ||=
-                      Types::Function::Param.new(
-                        type: untyped,
-                        name: nil
-                      )
-                  end
-                end
-              end
-            end
-          end
-        end
-
-        method_block
-      end
-
-      def keyword_hash?(node)
-        if node
-          if node.type == :HASH
-            node.children[0].children.compact.each_slice(2).all? {|key, _|
-              key.type == :LIT && key.children[0].is_a?(Symbol)
-            }
-          end
-        end
-      end
-
-      def any_node?(node, nodes: [], &block)
-        if yield(node)
-          nodes << node
-        end
-
-        each_child node do |child|
-          any_node? child, nodes: nodes, &block
-        end
-
-        nodes.empty? ? nil : nodes
-      end
-
-      def node_type(node, default: Types::Bases::Any.new(location: nil))
+      def param_type(node, default: Types::Bases::Any.new(location: nil))
         case node.type
         when :LIT
           case node.children[0]
@@ -664,9 +671,8 @@ module RBS
         end
       end
 
-      def untyped
-        @untyped ||= Types::Bases::Any.new(location: nil)
-      end
+      # backward compatible
+      alias node_type param_type
 
       def private
         @private ||= AST::Members::Private.new(location: nil)
@@ -677,11 +683,17 @@ module RBS
       end
 
       def current_accessibility(decls, index = decls.size)
-        idx = decls.slice(0, index).rindex { |decl| decl == private || decl == public }
-        (idx && decls[idx]) || public
+        slice = decls.slice(0, index) or raise
+        idx = slice.rindex { |decl| decl == private || decl == public }
+        if idx
+          _ = decls[idx]
+        else
+          public
+        end
       end
 
       def remove_unnecessary_accessibility_methods!(decls)
+        # @type var current: decl
         current = public
         idx = 0
 
@@ -703,7 +715,7 @@ module RBS
           idx += 1
         end
 
-        decls.pop while decls.last && is_accessibility?(decls.last)
+        decls.pop while decls.last && is_accessibility?(decls.last || raise)
       end
 
       def is_accessibility?(decl)
@@ -711,14 +723,21 @@ module RBS
       end
 
       def find_def_index_by_name(decls, name)
-        decls.find_index do |decl|
+        index = decls.find_index do |decl|
           case decl
           when AST::Members::MethodDefinition, AST::Members::AttrReader
             decl.name == name
           when AST::Members::AttrWriter
-            decl.name == :"#{name}="
+            :"#{decl.name}=" == name
           end
         end
+
+        if index
+          [
+            index,
+            _ = decls[index]
+          ]
+        end
       end
     end
   end
