@@ -26,7 +26,7 @@ rsock_syserr_fail_host_port(int err, const char *mesg, VALUE host, VALUE port)
     VALUE message;
 
     message = rb_sprintf("%s for %+"PRIsVALUE" port % "PRIsVALUE"",
-			 mesg, host, port);
+                         mesg, host, port);
 
     rb_syserr_fail_str(err, message);
 }
@@ -43,11 +43,11 @@ rsock_syserr_fail_path(int err, const char *mesg, VALUE path)
     VALUE message;
 
     if (RB_TYPE_P(path, T_STRING)) {
-	message = rb_sprintf("%s for % "PRIsVALUE"", mesg, path);
-	rb_syserr_fail_str(err, message);
+        message = rb_sprintf("%s for % "PRIsVALUE"", mesg, path);
+        rb_syserr_fail_str(err, message);
     }
     else {
-	rb_syserr_fail(err, mesg);
+        rb_syserr_fail(err, mesg);
     }
 }
 
@@ -96,12 +96,12 @@ rsock_syserr_fail_raddrinfo_or_sockaddr(int err, const char *mesg, VALUE addr, V
     if (NIL_P(rai)) {
         StringValue(addr);
 
-	rsock_syserr_fail_sockaddr(err, mesg,
+        rsock_syserr_fail_sockaddr(err, mesg,
             (struct sockaddr *)RSTRING_PTR(addr),
             (socklen_t)RSTRING_LEN(addr)); /* overflow should be checked already */
     }
     else
-	rsock_syserr_fail_raddrinfo(err, mesg, rai);
+        rsock_syserr_fail_raddrinfo(err, mesg, rai);
 }
 
 static void
@@ -168,93 +168,47 @@ pair_yield(VALUE pair)
 #endif
 
 #if defined HAVE_SOCKETPAIR
-
+static int
+rsock_socketpair0(int domain, int type, int protocol, int descriptors[2])
+{
 #ifdef SOCK_CLOEXEC
-static int
-rsock_socketpair0(int domain, int type, int protocol, int sv[2])
-{
-    int ret;
-    static int cloexec_state = -1; /* <0: unknown, 0: ignored, >0: working */
-    static const int default_flags = SOCK_CLOEXEC|RSOCK_NONBLOCK_DEFAULT;
+    type |= SOCK_CLOEXEC;
+#endif
 
-    if (cloexec_state > 0) { /* common path, if SOCK_CLOEXEC is defined */
-        ret = socketpair(domain, type|default_flags, protocol, sv);
-        if (ret == 0 && (sv[0] <= 2 || sv[1] <= 2)) {
-            goto fix_cloexec; /* highly unlikely */
-        }
-        goto update_max_fd;
-    }
-    else if (cloexec_state < 0) { /* usually runs once only for detection */
-        ret = socketpair(domain, type|default_flags, protocol, sv);
-        if (ret == 0) {
-            cloexec_state = rsock_detect_cloexec(sv[0]);
-            if ((cloexec_state == 0) || (sv[0] <= 2 || sv[1] <= 2))
-                goto fix_cloexec;
-            goto update_max_fd;
-        }
-        else if (ret == -1 && errno == EINVAL) {
-            /* SOCK_CLOEXEC is available since Linux 2.6.27.  Linux 2.6.18 fails with EINVAL */
-            ret = socketpair(domain, type, protocol, sv);
-            if (ret != -1) {
-                /* The reason of EINVAL may be other than SOCK_CLOEXEC.
-                 * So disable SOCK_CLOEXEC only if socketpair() succeeds without SOCK_CLOEXEC.
-                 * Ex. Socket.pair(:UNIX, 0xff) fails with EINVAL.
-                 */
-                cloexec_state = 0;
-            }
-        }
-    }
-    else { /* cloexec_state == 0 */
-        ret = socketpair(domain, type, protocol, sv);
-    }
-    if (ret == -1) {
+#ifdef SOCK_NONBLOCK
+    type |= SOCK_NONBLOCK;
+#endif
+
+    int result = socketpair(domain, type, protocol, descriptors);
+
+    if (result == -1)
         return -1;
-    }
 
-fix_cloexec:
-    rb_maygvl_fd_fix_cloexec(sv[0]);
-    rb_maygvl_fd_fix_cloexec(sv[1]);
-    if (RSOCK_NONBLOCK_DEFAULT) {
-        rsock_make_fd_nonblock(sv[0]);
-        rsock_make_fd_nonblock(sv[1]);
-    }
+#ifndef SOCK_CLOEXEC
+    rb_fd_fix_cloexec(descriptors[0]);
+    rb_fd_fix_cloexec(descriptors[1]);
+#endif
 
-update_max_fd:
-    rb_update_max_fd(sv[0]);
-    rb_update_max_fd(sv[1]);
+#ifndef SOCK_NONBLOCK
+    rsock_make_fd_nonblock(descriptors[0]);
+    rsock_make_fd_nonblock(descriptors[1]);
+#endif
 
-    return ret;
+    return result;
 }
-#else /* !SOCK_CLOEXEC */
-static int
-rsock_socketpair0(int domain, int type, int protocol, int sv[2])
-{
-    int ret = socketpair(domain, type, protocol, sv);
-
-    if (ret == -1)
-	return -1;
-
-    rb_fd_fix_cloexec(sv[0]);
-    rb_fd_fix_cloexec(sv[1]);
-    if (RSOCK_NONBLOCK_DEFAULT) {
-        rsock_make_fd_nonblock(sv[0]);
-        rsock_make_fd_nonblock(sv[1]);
-    }
-    return ret;
-}
-#endif /* !SOCK_CLOEXEC */
 
 static int
-rsock_socketpair(int domain, int type, int protocol, int sv[2])
+rsock_socketpair(int domain, int type, int protocol, int descriptors[2])
 {
-    int ret;
+    int result;
 
-    ret = rsock_socketpair0(domain, type, protocol, sv);
-    if (ret < 0 && rb_gc_for_fd(errno)) {
-        ret = rsock_socketpair0(domain, type, protocol, sv);
+    result = rsock_socketpair0(domain, type, protocol, descriptors);
+
+    if (result < 0 && rb_gc_for_fd(errno)) {
+        result = rsock_socketpair0(domain, type, protocol, descriptors);
     }
 
-    return ret;
+    return result;
 }
 
 /*
@@ -302,7 +256,7 @@ rsock_sock_s_socketpair(int argc, VALUE *argv, VALUE klass)
     p = NUM2INT(protocol);
     ret = rsock_socketpair(d, t, p, sp);
     if (ret < 0) {
-	rb_sys_fail("socketpair(2)");
+        rb_sys_fail("socketpair(2)");
     }
 
     s1 = rsock_init_sock(rb_obj_alloc(klass), sp[0]);
@@ -376,7 +330,7 @@ rsock_sock_s_socketpair(int argc, VALUE *argv, VALUE klass)
  * * Errno::EOPNOTSUPP - the calling +socket+ is listening and cannot be connected
  * * Errno::EPROTOTYPE - the _sockaddr_ has a different type than the socket
  *   bound to the specified peer address
- * * Errno::ETIMEDOUT - the attempt to connect time out before a connection
+ * * Errno::ETIMEDOUT - the attempt to connect timed out before a connection
  *   was made.
  *
  * On unix-based systems if the address family of the calling +socket+ is
@@ -417,7 +371,7 @@ rsock_sock_s_socketpair(int argc, VALUE *argv, VALUE klass)
  * * Errno::EHOSTUNREACH - no route to the network is present
  * * Errno::ENOBUFS - no buffer space is available
  * * Errno::ENOTSOCK - the +socket+ argument does not refer to a socket
- * * Errno::ETIMEDOUT - the attempt to connect time out before a connection
+ * * Errno::ETIMEDOUT - the attempt to connect timed out before a connection
  *   was made.
  * * Errno::EWOULDBLOCK - the socket is marked as nonblocking and the
  *   connection cannot be completed immediately
@@ -439,9 +393,9 @@ sock_connect(VALUE sock, VALUE addr)
     addr = rb_str_new4(addr);
     GetOpenFile(sock, fptr);
     fd = fptr->fd;
-    n = rsock_connect(fd, (struct sockaddr*)RSTRING_PTR(addr), RSTRING_SOCKLEN(addr), 0);
+    n = rsock_connect(fd, (struct sockaddr*)RSTRING_PTR(addr), RSTRING_SOCKLEN(addr), 0, NULL);
     if (n < 0) {
-	rsock_sys_fail_raddrinfo_or_sockaddr("connect(2)", addr, rai);
+        rsock_sys_fail_raddrinfo_or_sockaddr("connect(2)", addr, rai);
     }
 
     return INT2FIX(n);
@@ -461,19 +415,19 @@ sock_connect_nonblock(VALUE sock, VALUE addr, VALUE ex)
     rb_io_set_nonblock(fptr);
     n = connect(fptr->fd, (struct sockaddr*)RSTRING_PTR(addr), RSTRING_SOCKLEN(addr));
     if (n < 0) {
-	int e = errno;
-	if (e == EINPROGRESS) {
+        int e = errno;
+        if (e == EINPROGRESS) {
             if (ex == Qfalse) {
                 return sym_wait_writable;
             }
             rb_readwrite_syserr_fail(RB_IO_WAIT_WRITABLE, e, "connect(2) would block");
-	}
-	if (e == EISCONN) {
+        }
+        if (e == EISCONN) {
             if (ex == Qfalse) {
                 return INT2FIX(0);
             }
-	}
-	rsock_syserr_fail_raddrinfo_or_sockaddr(e, "connect(2)", addr, rai);
+        }
+        rsock_syserr_fail_raddrinfo_or_sockaddr(e, "connect(2)", addr, rai);
     }
 
     return INT2FIX(n);
@@ -574,7 +528,7 @@ sock_bind(VALUE sock, VALUE addr)
     SockAddrStringValueWithAddrinfo(addr, rai);
     GetOpenFile(sock, fptr);
     if (bind(fptr->fd, (struct sockaddr*)RSTRING_PTR(addr), RSTRING_SOCKLEN(addr)) < 0)
-	rsock_sys_fail_raddrinfo_or_sockaddr("bind(2)", addr, rai);
+        rsock_sys_fail_raddrinfo_or_sockaddr("bind(2)", addr, rai);
 
     return INT2FIX(0);
 }
@@ -658,7 +612,7 @@ rsock_sock_listen(VALUE sock, VALUE log)
     backlog = NUM2INT(log);
     GetOpenFile(sock, fptr);
     if (listen(fptr->fd, backlog) < 0)
-	rb_sys_fail("listen(2)");
+        rb_sys_fail("listen(2)");
 
     return INT2FIX(0);
 }
@@ -796,17 +750,14 @@ sock_recvfrom_nonblock(VALUE sock, VALUE len, VALUE flg, VALUE str, VALUE ex)
  *
  */
 static VALUE
-sock_accept(VALUE sock)
+sock_accept(VALUE server)
 {
-    rb_io_t *fptr;
-    VALUE sock2;
-    union_sockaddr buf;
-    socklen_t len = (socklen_t)sizeof buf;
+    union_sockaddr buffer;
+    socklen_t length = (socklen_t)sizeof(buffer);
 
-    GetOpenFile(sock, fptr);
-    sock2 = rsock_s_accept(rb_cSocket,fptr->fd,&buf.addr,&len);
+    VALUE peer = rsock_s_accept(rb_cSocket, server, &buffer.addr, &length);
 
-    return rb_assoc_new(sock2, rsock_io_socket_addrinfo(sock2, &buf.addr, len));
+    return rb_assoc_new(peer, rsock_io_socket_addrinfo(peer, &buffer.addr, length));
 }
 
 /* :nodoc: */
@@ -823,7 +774,7 @@ sock_accept_nonblock(VALUE sock, VALUE ex)
     sock2 = rsock_s_accept_nonblock(rb_cSocket, ex, fptr, addr, &len);
 
     if (SYMBOL_P(sock2)) /* :wait_readable */
-	return sock2;
+        return sock2;
     return rb_assoc_new(sock2, rsock_io_socket_addrinfo(sock2, &buf.addr, len));
 }
 
@@ -866,17 +817,14 @@ sock_accept_nonblock(VALUE sock, VALUE ex)
  * * Socket#accept
  */
 static VALUE
-sock_sysaccept(VALUE sock)
+sock_sysaccept(VALUE server)
 {
-    rb_io_t *fptr;
-    VALUE sock2;
-    union_sockaddr buf;
-    socklen_t len = (socklen_t)sizeof buf;
+    union_sockaddr buffer;
+    socklen_t length = (socklen_t)sizeof(buffer);
 
-    GetOpenFile(sock, fptr);
-    sock2 = rsock_s_accept(0,fptr->fd,&buf.addr,&len);
+    VALUE peer = rsock_s_accept(0, server, &buffer.addr, &length);
 
-    return rb_assoc_new(sock2, rsock_io_socket_addrinfo(sock2, &buf.addr, len));
+    return rb_assoc_new(peer, rsock_io_socket_addrinfo(peer, &buffer.addr, length));
 }
 
 #ifdef HAVE_GETHOSTNAME
@@ -907,19 +855,19 @@ sock_gethostname(VALUE obj)
 
     name = rb_str_new(0, len);
     while (gethostname(RSTRING_PTR(name), len) < 0) {
-	int e = errno;
-	switch (e) {
-	  case ENAMETOOLONG:
+        int e = errno;
+        switch (e) {
+          case ENAMETOOLONG:
 #ifdef __linux__
-	  case EINVAL:
-	    /* glibc before version 2.1 uses EINVAL instead of ENAMETOOLONG */
+          case EINVAL:
+            /* glibc before version 2.1 uses EINVAL instead of ENAMETOOLONG */
 #endif
-	    break;
-	  default:
-	    rb_syserr_fail(e, "gethostname(3)");
-	}
-	rb_str_modify_expand(name, len);
-	len += len;
+            break;
+          default:
+            rb_syserr_fail(e, "gethostname(3)");
+        }
+        rb_str_modify_expand(name, len);
+        len += len;
     }
     rb_str_resize(name, strlen(RSTRING_PTR(name)));
     return name;
@@ -949,18 +897,18 @@ make_addrinfo(struct rb_addrinfo *res0, int norevlookup)
     struct addrinfo *res;
 
     if (res0 == NULL) {
-	rb_raise(rb_eSocket, "host not found");
+        rb_raise(rb_eSocket, "host not found");
     }
     base = rb_ary_new();
     for (res = res0->ai; res; res = res->ai_next) {
-	ary = rsock_ipaddr(res->ai_addr, res->ai_addrlen, norevlookup);
-	if (res->ai_canonname) {
-	    RARRAY_ASET(ary, 2, rb_str_new2(res->ai_canonname));
-	}
-	rb_ary_push(ary, INT2FIX(res->ai_family));
-	rb_ary_push(ary, INT2FIX(res->ai_socktype));
-	rb_ary_push(ary, INT2FIX(res->ai_protocol));
-	rb_ary_push(base, ary);
+        ary = rsock_ipaddr(res->ai_addr, res->ai_addrlen, norevlookup);
+        if (res->ai_canonname) {
+            RARRAY_ASET(ary, 2, rb_str_new2(res->ai_canonname));
+        }
+        rb_ary_push(ary, INT2FIX(res->ai_family));
+        rb_ary_push(ary, INT2FIX(res->ai_socktype));
+        rb_ary_push(ary, INT2FIX(res->ai_protocol));
+        rb_ary_push(base, ary);
     }
     return base;
 }
@@ -972,18 +920,18 @@ sock_sockaddr(struct sockaddr *addr, socklen_t len)
 
     switch (addr->sa_family) {
       case AF_INET:
-	ptr = (char*)&((struct sockaddr_in*)addr)->sin_addr.s_addr;
-	len = (socklen_t)sizeof(((struct sockaddr_in*)addr)->sin_addr.s_addr);
-	break;
+        ptr = (char*)&((struct sockaddr_in*)addr)->sin_addr.s_addr;
+        len = (socklen_t)sizeof(((struct sockaddr_in*)addr)->sin_addr.s_addr);
+        break;
 #ifdef AF_INET6
       case AF_INET6:
-	ptr = (char*)&((struct sockaddr_in6*)addr)->sin6_addr.s6_addr;
-	len = (socklen_t)sizeof(((struct sockaddr_in6*)addr)->sin6_addr.s6_addr);
-	break;
+        ptr = (char*)&((struct sockaddr_in6*)addr)->sin6_addr.s6_addr;
+        len = (socklen_t)sizeof(((struct sockaddr_in6*)addr)->sin6_addr.s6_addr);
+        break;
 #endif
       default:
         rb_raise(rb_eSocket, "unknown socket family:%d", addr->sa_family);
-	break;
+        break;
     }
     return rb_str_new(ptr, len);
 }
@@ -1011,8 +959,9 @@ sock_sockaddr(struct sockaddr *addr, socklen_t len)
 static VALUE
 sock_s_gethostbyname(VALUE obj, VALUE host)
 {
+    rb_warn("Socket.gethostbyname is deprecated; use Addrinfo.getaddrinfo instead.");
     struct rb_addrinfo *res =
-	rsock_addrinfo(host, Qnil, AF_UNSPEC, SOCK_STREAM, AI_CANONNAME);
+        rsock_addrinfo(host, Qnil, AF_UNSPEC, SOCK_STREAM, AI_CANONNAME);
     return rsock_make_hostent(host, res, sock_sockaddr);
 }
 
@@ -1050,23 +999,25 @@ sock_s_gethostbyaddr(int argc, VALUE *argv, VALUE _)
     VALUE ary, names;
     int t = AF_INET;
 
+    rb_warn("Socket.gethostbyaddr is deprecated; use Addrinfo#getnameinfo instead.");
+
     rb_scan_args(argc, argv, "11", &addr, &family);
     StringValue(addr);
     if (!NIL_P(family)) {
-	t = rsock_family_arg(family);
+        t = rsock_family_arg(family);
     }
 #ifdef AF_INET6
     else if (RSTRING_LEN(addr) == 16) {
-	t = AF_INET6;
+        t = AF_INET6;
     }
 #endif
     h = gethostbyaddr(RSTRING_PTR(addr), RSTRING_SOCKLEN(addr), t);
     if (h == NULL) {
 #ifdef HAVE_HSTRERROR
-	extern int h_errno;
-	rb_raise(rb_eSocket, "%s", (char*)hstrerror(h_errno));
+        extern int h_errno;
+        rb_raise(rb_eSocket, "%s", (char*)hstrerror(h_errno));
 #else
-	rb_raise(rb_eSocket, "host not found");
+        rb_raise(rb_eSocket, "host not found");
 #endif
     }
     ary = rb_ary_new();
@@ -1074,14 +1025,14 @@ sock_s_gethostbyaddr(int argc, VALUE *argv, VALUE _)
     names = rb_ary_new();
     rb_ary_push(ary, names);
     if (h->h_aliases != NULL) {
-	for (pch = h->h_aliases; *pch; pch++) {
-	    rb_ary_push(names, rb_str_new2(*pch));
-	}
+        for (pch = h->h_aliases; *pch; pch++) {
+            rb_ary_push(names, rb_str_new2(*pch));
+        }
     }
     rb_ary_push(ary, INT2NUM(h->h_addrtype));
 #ifdef h_addr
     for (pch = h->h_addr_list; *pch; pch++) {
-	rb_ary_push(ary, rb_str_new(*pch, h->h_length));
+        rb_ary_push(ary, rb_str_new(*pch, h->h_length));
     }
 #else
     rb_ary_push(ary, rb_str_new(h->h_addr, h->h_length));
@@ -1118,15 +1069,15 @@ sock_s_getservbyname(int argc, VALUE *argv, VALUE _)
     if (!NIL_P(proto)) protoname = StringValueCStr(proto);
     sp = getservbyname(servicename, protoname);
     if (sp) {
-	port = ntohs(sp->s_port);
+        port = ntohs(sp->s_port);
     }
     else {
-	char *end;
+        char *end;
 
-	port = STRTOUL(servicename, &end, 0);
-	if (*end != '\0') {
-	    rb_raise(rb_eSocket, "no such service %s/%s", servicename, protoname);
-	}
+        port = STRTOUL(servicename, &end, 0);
+        if (*end != '\0') {
+            rb_raise(rb_eSocket, "no such service %s/%s", servicename, protoname);
+        }
     }
     return INT2FIX(port);
 }
@@ -1155,16 +1106,16 @@ sock_s_getservbyport(int argc, VALUE *argv, VALUE _)
     rb_scan_args(argc, argv, "11", &port, &proto);
     portnum = NUM2LONG(port);
     if (portnum != (uint16_t)portnum) {
-	const char *s = portnum > 0 ? "big" : "small";
-	rb_raise(rb_eRangeError, "integer %ld too %s to convert into `int16_t'", portnum, s);
+        const char *s = portnum > 0 ? "big" : "small";
+        rb_raise(rb_eRangeError, "integer %ld too %s to convert into `int16_t'", portnum, s);
     }
     if (!NIL_P(proto)) protoname = StringValueCStr(proto);
 
     sp = getservbyport((int)htons((uint16_t)portnum), protoname);
     if (!sp) {
-	rb_raise(rb_eSocket, "no such service for port %d/%s", (int)portnum, protoname);
+        rb_raise(rb_eSocket, "no such service for port %d/%s", (int)portnum, protoname);
     }
-    return rb_tainted_str_new2(sp->s_name);
+    return rb_str_new2(sp->s_name);
 }
 
 /*
@@ -1197,7 +1148,7 @@ sock_s_getservbyport(int argc, VALUE *argv, VALUE _)
  * be one of below.  If _reverse_lookup_ is omitted, the default value is +nil+.
  *
  *   +true+, +:hostname+:  hostname is obtained from numeric address using reverse lookup, which may take a time.
- *   +false+, +:numeric+:  hostname is same as numeric address.
+ *   +false+, +:numeric+:  hostname is the same as numeric address.
  *   +nil+:              obey to the current +do_not_reverse_lookup+ flag.
  *
  * If Addrinfo object is preferred, use Addrinfo.getaddrinfo.
@@ -1216,17 +1167,18 @@ sock_s_getaddrinfo(int argc, VALUE *argv, VALUE _)
     hints.ai_family = NIL_P(family) ? PF_UNSPEC : rsock_family_arg(family);
 
     if (!NIL_P(socktype)) {
-	hints.ai_socktype = rsock_socktype_arg(socktype);
+        hints.ai_socktype = rsock_socktype_arg(socktype);
     }
     if (!NIL_P(protocol)) {
-	hints.ai_protocol = NUM2INT(protocol);
+        hints.ai_protocol = NUM2INT(protocol);
     }
     if (!NIL_P(flags)) {
-	hints.ai_flags = NUM2INT(flags);
+        hints.ai_flags = NUM2INT(flags);
     }
     if (NIL_P(revlookup) || !rsock_revlookup_flag(revlookup, &norevlookup)) {
-	norevlookup = rsock_do_not_reverse_lookup;
+        norevlookup = rsock_do_not_reverse_lookup;
     }
+
     res = rsock_getaddrinfo(host, port, &hints, 0);
 
     ret = make_addrinfo(res, norevlookup);
@@ -1260,7 +1212,6 @@ static VALUE
 sock_s_getnameinfo(int argc, VALUE *argv, VALUE _)
 {
     VALUE sa, af = Qnil, host = Qnil, port = Qnil, flags, tmp;
-    char *hptr, *pptr;
     char hbuf[1024], pbuf[1024];
     int fl;
     struct rb_addrinfo *res = NULL;
@@ -1275,115 +1226,85 @@ sock_s_getnameinfo(int argc, VALUE *argv, VALUE _)
 
     fl = 0;
     if (!NIL_P(flags)) {
-	fl = NUM2INT(flags);
+        fl = NUM2INT(flags);
     }
     tmp = rb_check_sockaddr_string_type(sa);
     if (!NIL_P(tmp)) {
-	sa = tmp;
-	if (sizeof(ss) < (size_t)RSTRING_LEN(sa)) {
-	    rb_raise(rb_eTypeError, "sockaddr length too big");
-	}
-	memcpy(&ss, RSTRING_PTR(sa), RSTRING_LEN(sa));
+        sa = tmp;
+        if (sizeof(ss) < (size_t)RSTRING_LEN(sa)) {
+            rb_raise(rb_eTypeError, "sockaddr length too big");
+        }
+        memcpy(&ss, RSTRING_PTR(sa), RSTRING_LEN(sa));
         if (!VALIDATE_SOCKLEN(&ss.addr, RSTRING_LEN(sa))) {
-	    rb_raise(rb_eTypeError, "sockaddr size differs - should not happen");
-	}
-	sap = &ss.addr;
+            rb_raise(rb_eTypeError, "sockaddr size differs - should not happen");
+        }
+        sap = &ss.addr;
         salen = RSTRING_SOCKLEN(sa);
-	goto call_nameinfo;
+        goto call_nameinfo;
     }
     tmp = rb_check_array_type(sa);
     if (!NIL_P(tmp)) {
-	sa = tmp;
-	MEMZERO(&hints, struct addrinfo, 1);
-	if (RARRAY_LEN(sa) == 3) {
-	    af = RARRAY_AREF(sa, 0);
-	    port = RARRAY_AREF(sa, 1);
-	    host = RARRAY_AREF(sa, 2);
-	}
-	else if (RARRAY_LEN(sa) >= 4) {
-	    af = RARRAY_AREF(sa, 0);
-	    port = RARRAY_AREF(sa, 1);
-	    host = RARRAY_AREF(sa, 3);
-	    if (NIL_P(host)) {
-		host = RARRAY_AREF(sa, 2);
-	    }
-	    else {
-		/*
-		 * 4th element holds numeric form, don't resolve.
-		 * see rsock_ipaddr().
-		 */
+        sa = tmp;
+        MEMZERO(&hints, struct addrinfo, 1);
+        if (RARRAY_LEN(sa) == 3) {
+            af = RARRAY_AREF(sa, 0);
+            port = RARRAY_AREF(sa, 1);
+            host = RARRAY_AREF(sa, 2);
+        }
+        else if (RARRAY_LEN(sa) >= 4) {
+            af = RARRAY_AREF(sa, 0);
+            port = RARRAY_AREF(sa, 1);
+            host = RARRAY_AREF(sa, 3);
+            if (NIL_P(host)) {
+                host = RARRAY_AREF(sa, 2);
+            }
+            else {
+                /*
+                 * 4th element holds numeric form, don't resolve.
+                 * see rsock_ipaddr().
+                 */
 #ifdef AI_NUMERICHOST /* AIX 4.3.3 doesn't have AI_NUMERICHOST. */
-		hints.ai_flags |= AI_NUMERICHOST;
+                hints.ai_flags |= AI_NUMERICHOST;
 #endif
-	    }
-	}
-	else {
-	    rb_raise(rb_eArgError, "array size should be 3 or 4, %ld given",
-		     RARRAY_LEN(sa));
-	}
-	/* host */
-	if (NIL_P(host)) {
-	    hptr = NULL;
-	}
-	else {
-	    strncpy(hbuf, StringValueCStr(host), sizeof(hbuf));
-	    hbuf[sizeof(hbuf) - 1] = '\0';
-	    hptr = hbuf;
-	}
-	/* port */
-	if (NIL_P(port)) {
-	    strcpy(pbuf, "0");
-	    pptr = NULL;
-	}
-	else if (FIXNUM_P(port)) {
-	    snprintf(pbuf, sizeof(pbuf), "%ld", NUM2LONG(port));
-	    pptr = pbuf;
-	}
-	else {
-	    strncpy(pbuf, StringValueCStr(port), sizeof(pbuf));
-	    pbuf[sizeof(pbuf) - 1] = '\0';
-	    pptr = pbuf;
-	}
-	hints.ai_socktype = (fl & NI_DGRAM) ? SOCK_DGRAM : SOCK_STREAM;
-	/* af */
+            }
+        }
+        else {
+            rb_raise(rb_eArgError, "array size should be 3 or 4, %ld given",
+                     RARRAY_LEN(sa));
+        }
+        hints.ai_socktype = (fl & NI_DGRAM) ? SOCK_DGRAM : SOCK_STREAM;
+        /* af */
         hints.ai_family = NIL_P(af) ? PF_UNSPEC : rsock_family_arg(af);
-	error = rb_getaddrinfo(hptr, pptr, &hints, &res);
-	if (error) goto error_exit_addr;
-	sap = res->ai->ai_addr;
+        res = rsock_getaddrinfo(host, port, &hints, 0);
+        sap = res->ai->ai_addr;
         salen = res->ai->ai_addrlen;
     }
     else {
-	rb_raise(rb_eTypeError, "expecting String or Array");
+        rb_raise(rb_eTypeError, "expecting String or Array");
     }
 
   call_nameinfo:
     error = rb_getnameinfo(sap, salen, hbuf, sizeof(hbuf),
-			   pbuf, sizeof(pbuf), fl);
+                           pbuf, sizeof(pbuf), fl);
     if (error) goto error_exit_name;
     if (res) {
-	for (r = res->ai->ai_next; r; r = r->ai_next) {
-	    char hbuf2[1024], pbuf2[1024];
+        for (r = res->ai->ai_next; r; r = r->ai_next) {
+            char hbuf2[1024], pbuf2[1024];
 
-	    sap = r->ai_addr;
+            sap = r->ai_addr;
             salen = r->ai_addrlen;
-	    error = rb_getnameinfo(sap, salen, hbuf2, sizeof(hbuf2),
-				   pbuf2, sizeof(pbuf2), fl);
-	    if (error) goto error_exit_name;
-	    if (strcmp(hbuf, hbuf2) != 0|| strcmp(pbuf, pbuf2) != 0) {
-		rb_freeaddrinfo(res);
-		rb_raise(rb_eSocket, "sockaddr resolved to multiple nodename");
-	    }
-	}
-	rb_freeaddrinfo(res);
+            error = rb_getnameinfo(sap, salen, hbuf2, sizeof(hbuf2),
+                                   pbuf2, sizeof(pbuf2), fl);
+            if (error) goto error_exit_name;
+            if (strcmp(hbuf, hbuf2) != 0|| strcmp(pbuf, pbuf2) != 0) {
+                rb_freeaddrinfo(res);
+                rb_raise(rb_eSocket, "sockaddr resolved to multiple nodename");
+            }
+        }
+        rb_freeaddrinfo(res);
     }
     return rb_assoc_new(rb_str_new2(hbuf), rb_str_new2(pbuf));
 
-  error_exit_addr:
-    saved_errno = errno;
-    if (res) rb_freeaddrinfo(res);
-    errno = saved_errno;
-    rsock_raise_socket_error("getaddrinfo", error);
-
   error_exit_name:
     saved_errno = errno;
     if (res) rb_freeaddrinfo(res);
@@ -1414,8 +1335,6 @@ sock_s_pack_sockaddr_in(VALUE self, VALUE port, VALUE host)
     VALUE addr = rb_str_new((char*)res->ai->ai_addr, res->ai->ai_addrlen);
 
     rb_freeaddrinfo(res);
-    OBJ_INFECT(addr, port);
-    OBJ_INFECT(addr, host);
 
     return addr;
 }
@@ -1457,7 +1376,6 @@ sock_s_unpack_sockaddr_in(VALUE self, VALUE addr)
 #endif
     }
     host = rsock_make_ipaddr((struct sockaddr*)sockaddr, RSTRING_SOCKLEN(addr));
-    OBJ_INFECT(host, addr);
     return rb_assoc_new(INT2NUM(ntohs(sockaddr->sin_port)), host);
 }
 
@@ -1487,7 +1405,6 @@ sock_s_pack_sockaddr_un(VALUE self, VALUE path)
     }
     memcpy(sockaddr.sun_path, RSTRING_PTR(path), RSTRING_LEN(path));
     addr = rb_str_new((char*)&sockaddr, rsock_unix_sockaddr_len(path));
-    OBJ_INFECT(addr, path);
 
     return addr;
 }
@@ -1520,11 +1437,10 @@ sock_s_unpack_sockaddr_un(VALUE self, VALUE addr)
         rb_raise(rb_eArgError, "not an AF_UNIX sockaddr");
     }
     if (sizeof(struct sockaddr_un) < (size_t)RSTRING_LEN(addr)) {
-	rb_raise(rb_eTypeError, "too long sockaddr_un - %ld longer than %d",
-		 RSTRING_LEN(addr), (int)sizeof(struct sockaddr_un));
+        rb_raise(rb_eTypeError, "too long sockaddr_un - %ld longer than %d",
+                 RSTRING_LEN(addr), (int)sizeof(struct sockaddr_un));
     }
     path = rsock_unixpath_str(sockaddr, RSTRING_SOCKLEN(addr));
-    OBJ_INFECT(path, addr);
     return path;
 }
 #endif
@@ -1586,19 +1502,19 @@ sockaddr_obj(struct sockaddr *addr, socklen_t len)
 
 #if defined(__KAME__) && defined(AF_INET6)
     if (addr->sa_family == AF_INET6) {
-	/* KAME uses the 2nd 16bit word of link local IPv6 address as interface index internally */
+        /* KAME uses the 2nd 16bit word of link local IPv6 address as interface index internally */
         /* http://orange.kame.net/dev/cvsweb.cgi/kame/IMPLEMENTATION */
-	/* convert fe80:1::1 to fe80::1%1 */
+        /* convert fe80:1::1 to fe80::1%1 */
         len = (socklen_t)sizeof(struct sockaddr_in6);
-	memcpy(&addr6, addr, len);
-	addr = (struct sockaddr *)&addr6;
-	if (IN6_IS_ADDR_LINKLOCAL(&addr6.sin6_addr) &&
-	    addr6.sin6_scope_id == 0 &&
-	    (addr6.sin6_addr.s6_addr[2] || addr6.sin6_addr.s6_addr[3])) {
-	    addr6.sin6_scope_id = (addr6.sin6_addr.s6_addr[2] << 8) | addr6.sin6_addr.s6_addr[3];
-	    addr6.sin6_addr.s6_addr[2] = 0;
-	    addr6.sin6_addr.s6_addr[3] = 0;
-	}
+        memcpy(&addr6, addr, len);
+        addr = (struct sockaddr *)&addr6;
+        if (IN6_IS_ADDR_LINKLOCAL(&addr6.sin6_addr) &&
+            addr6.sin6_scope_id == 0 &&
+            (addr6.sin6_addr.s6_addr[2] || addr6.sin6_addr.s6_addr[3])) {
+            addr6.sin6_scope_id = (addr6.sin6_addr.s6_addr[2] << 8) | addr6.sin6_addr.s6_addr[3];
+            addr6.sin6_addr.s6_addr[2] = 0;
+            addr6.sin6_addr.s6_addr[3] = 0;
+        }
     }
 #endif
 
@@ -1613,7 +1529,7 @@ rsock_sockaddr_obj(struct sockaddr *addr, socklen_t len)
 
 #endif
 
-#if defined(HAVE_GETIFADDRS) || (defined(SIOCGLIFCONF) && defined(SIOCGLIFNUM) && !defined(__hpux)) || defined(SIOCGIFCONF) ||  defined(_WIN32)
+#if defined(HAVE_GETIFADDRS) || (defined(SIOCGLIFCONF) && defined(SIOCGLIFNUM)) || defined(SIOCGIFCONF) ||  defined(_WIN32)
 /*
  * call-seq:
  *   Socket.ip_address_list => array
@@ -1674,9 +1590,8 @@ socket_s_ip_address_list(VALUE self)
     freeifaddrs(ifp);
 
     return list;
-#elif defined(SIOCGLIFCONF) && defined(SIOCGLIFNUM) && !defined(__hpux)
+#elif defined(SIOCGLIFCONF) && defined(SIOCGLIFNUM)
     /* Solaris if_tcp(7P) */
-    /* HP-UX has SIOCGLIFCONF too.  But it uses different struct */
     int fd = -1;
     int ret;
     struct lifnum ln;
@@ -1697,8 +1612,8 @@ socket_s_ip_address_list(VALUE self)
 
     ret = ioctl(fd, SIOCGLIFNUM, &ln);
     if (ret == -1) {
-	reason = "SIOCGLIFNUM";
-	goto finish;
+        reason = "SIOCGLIFNUM";
+        goto finish;
     }
 
     memset(&lc, 0, sizeof(lc));
@@ -1709,13 +1624,13 @@ socket_s_ip_address_list(VALUE self)
 
     ret = ioctl(fd, SIOCGLIFCONF, &lc);
     if (ret == -1) {
-	reason = "SIOCGLIFCONF";
-	goto finish;
+        reason = "SIOCGLIFCONF";
+        goto finish;
     }
 
     list = rb_ary_new();
     for (i = 0; i < ln.lifn_count; i++) {
-	struct lifreq *req = &lc.lifc_req[i];
+        struct lifreq *req = &lc.lifc_req[i];
         if (IS_IP_FAMILY(req->lifr_addr.ss_family)) {
             if (req->lifr_addr.ss_family == AF_INET6 &&
                 IN6_IS_ADDR_LINKLOCAL(&((struct sockaddr_in6 *)(&req->lifr_addr))->sin6_addr) &&
@@ -1736,13 +1651,13 @@ socket_s_ip_address_list(VALUE self)
   finish:
     save_errno = errno;
     if (lc.lifc_buf != NULL)
-	xfree(lc.lifc_req);
+        xfree(lc.lifc_req);
     if (fd != -1)
-	close(fd);
+        close(fd);
     errno = save_errno;
 
     if (reason)
-	rb_syserr_fail(save_errno, reason);
+        rb_syserr_fail(save_errno, reason);
     return list;
 
 #elif defined(SIOCGIFCONF)
@@ -1780,17 +1695,17 @@ socket_s_ip_address_list(VALUE self)
     /* fprintf(stderr, "conf.ifc_len: %d\n", conf.ifc_len); */
 
     if (bufsize - EXTRA_SPACE < conf.ifc_len) {
-	if (bufsize < conf.ifc_len) {
-	    /* NetBSD returns required size for all interfaces. */
-	    bufsize = conf.ifc_len + EXTRA_SPACE;
-	}
-	else {
-	    bufsize = bufsize << 1;
-	}
-	if (buf == initbuf)
-	    buf = NULL;
-	buf = xrealloc(buf, bufsize);
-	goto retry;
+        if (bufsize < conf.ifc_len) {
+            /* NetBSD returns required size for all interfaces. */
+            bufsize = conf.ifc_len + EXTRA_SPACE;
+        }
+        else {
+            bufsize = bufsize << 1;
+        }
+        if (buf == initbuf)
+            buf = NULL;
+        buf = xrealloc(buf, bufsize);
+        goto retry;
     }
 
     close(fd);
@@ -1799,10 +1714,10 @@ socket_s_ip_address_list(VALUE self)
     list = rb_ary_new();
     req = conf.ifc_req;
     while ((char*)req < (char*)conf.ifc_req + conf.ifc_len) {
-	struct sockaddr *addr = &req->ifr_addr;
+        struct sockaddr *addr = &req->ifr_addr;
         if (IS_IP_FAMILY(addr->sa_family)) {
-	    rb_ary_push(list, sockaddr_obj(addr, sockaddr_len(addr)));
-	}
+            rb_ary_push(list, sockaddr_obj(addr, sockaddr_len(addr)));
+        }
 #ifdef HAVE_STRUCT_SOCKADDR_SA_LEN
 # ifndef _SIZEOF_ADDR_IFREQ
 #  define _SIZEOF_ADDR_IFREQ(r) \
@@ -1811,9 +1726,9 @@ socket_s_ip_address_list(VALUE self)
             (r).ifr_addr.sa_len - sizeof(struct sockaddr) : \
             0))
 # endif
-	req = (struct ifreq *)((char*)req + _SIZEOF_ADDR_IFREQ(*req));
+        req = (struct ifreq *)((char*)req + _SIZEOF_ADDR_IFREQ(*req));
 #else
-	req = (struct ifreq *)((char*)req + sizeof(struct ifreq));
+        req = (struct ifreq *)((char*)req + sizeof(struct ifreq));
 #endif
     }
 
@@ -1823,57 +1738,57 @@ socket_s_ip_address_list(VALUE self)
     if (buf != initbuf)
         xfree(buf);
     if (fd != -1)
-	close(fd);
+        close(fd);
     errno = save_errno;
 
     if (reason)
-	rb_syserr_fail(save_errno, reason);
+        rb_syserr_fail(save_errno, reason);
     return list;
 
 #undef EXTRA_SPACE
 #elif defined(_WIN32)
     typedef struct ip_adapter_unicast_address_st {
-	unsigned LONG_LONG dummy0;
-	struct ip_adapter_unicast_address_st *Next;
-	struct {
-	    struct sockaddr *lpSockaddr;
-	    int iSockaddrLength;
-	} Address;
-	int dummy1;
-	int dummy2;
-	int dummy3;
-	long dummy4;
-	long dummy5;
-	long dummy6;
+        unsigned LONG_LONG dummy0;
+        struct ip_adapter_unicast_address_st *Next;
+        struct {
+            struct sockaddr *lpSockaddr;
+            int iSockaddrLength;
+        } Address;
+        int dummy1;
+        int dummy2;
+        int dummy3;
+        long dummy4;
+        long dummy5;
+        long dummy6;
     } ip_adapter_unicast_address_t;
     typedef struct ip_adapter_anycast_address_st {
-	unsigned LONG_LONG dummy0;
-	struct ip_adapter_anycast_address_st *Next;
-	struct {
-	    struct sockaddr *lpSockaddr;
-	    int iSockaddrLength;
-	} Address;
+        unsigned LONG_LONG dummy0;
+        struct ip_adapter_anycast_address_st *Next;
+        struct {
+            struct sockaddr *lpSockaddr;
+            int iSockaddrLength;
+        } Address;
     } ip_adapter_anycast_address_t;
     typedef struct ip_adapter_addresses_st {
-	unsigned LONG_LONG dummy0;
-	struct ip_adapter_addresses_st *Next;
-	void *dummy1;
-	ip_adapter_unicast_address_t *FirstUnicastAddress;
-	ip_adapter_anycast_address_t *FirstAnycastAddress;
-	void *dummy2;
-	void *dummy3;
-	void *dummy4;
-	void *dummy5;
-	void *dummy6;
-	BYTE dummy7[8];
-	DWORD dummy8;
-	DWORD dummy9;
-	DWORD dummy10;
-	DWORD IfType;
-	int OperStatus;
-	DWORD dummy12;
-	DWORD dummy13[16];
-	void *dummy14;
+        unsigned LONG_LONG dummy0;
+        struct ip_adapter_addresses_st *Next;
+        void *dummy1;
+        ip_adapter_unicast_address_t *FirstUnicastAddress;
+        ip_adapter_anycast_address_t *FirstAnycastAddress;
+        void *dummy2;
+        void *dummy3;
+        void *dummy4;
+        void *dummy5;
+        void *dummy6;
+        BYTE dummy7[8];
+        DWORD dummy8;
+        DWORD dummy9;
+        DWORD dummy10;
+        DWORD IfType;
+        int OperStatus;
+        DWORD dummy12;
+        DWORD dummy13[16];
+        void *dummy14;
     } ip_adapter_addresses_t;
     typedef ULONG (WINAPI *GetAdaptersAddresses_t)(ULONG, ULONG, PVOID, ip_adapter_addresses_t *, PULONG);
     HMODULE h;
@@ -1885,49 +1800,49 @@ socket_s_ip_address_list(VALUE self)
 
     h = LoadLibrary("iphlpapi.dll");
     if (!h)
-	rb_notimplement();
+        rb_notimplement();
     pGetAdaptersAddresses = (GetAdaptersAddresses_t)GetProcAddress(h, "GetAdaptersAddresses");
     if (!pGetAdaptersAddresses) {
-	FreeLibrary(h);
-	rb_notimplement();
+        FreeLibrary(h);
+        rb_notimplement();
     }
 
     ret = pGetAdaptersAddresses(AF_UNSPEC, 0, NULL, NULL, &len);
     if (ret != ERROR_SUCCESS && ret != ERROR_BUFFER_OVERFLOW) {
-	errno = rb_w32_map_errno(ret);
-	FreeLibrary(h);
-	rb_sys_fail("GetAdaptersAddresses");
+        errno = rb_w32_map_errno(ret);
+        FreeLibrary(h);
+        rb_sys_fail("GetAdaptersAddresses");
     }
     adapters = (ip_adapter_addresses_t *)ALLOCA_N(BYTE, len);
     ret = pGetAdaptersAddresses(AF_UNSPEC, 0, NULL, adapters, &len);
     if (ret != ERROR_SUCCESS) {
-	errno = rb_w32_map_errno(ret);
-	FreeLibrary(h);
-	rb_sys_fail("GetAdaptersAddresses");
+        errno = rb_w32_map_errno(ret);
+        FreeLibrary(h);
+        rb_sys_fail("GetAdaptersAddresses");
     }
 
     list = rb_ary_new();
     for (; adapters; adapters = adapters->Next) {
-	ip_adapter_unicast_address_t *uni;
-	ip_adapter_anycast_address_t *any;
-	if (adapters->OperStatus != 1)	/* 1 means IfOperStatusUp */
-	    continue;
-	for (uni = adapters->FirstUnicastAddress; uni; uni = uni->Next) {
+        ip_adapter_unicast_address_t *uni;
+        ip_adapter_anycast_address_t *any;
+        if (adapters->OperStatus != 1)	/* 1 means IfOperStatusUp */
+            continue;
+        for (uni = adapters->FirstUnicastAddress; uni; uni = uni->Next) {
 #ifndef INET6
-	    if (uni->Address.lpSockaddr->sa_family == AF_INET)
+            if (uni->Address.lpSockaddr->sa_family == AF_INET)
 #else
-	    if (IS_IP_FAMILY(uni->Address.lpSockaddr->sa_family))
+            if (IS_IP_FAMILY(uni->Address.lpSockaddr->sa_family))
 #endif
-		rb_ary_push(list, sockaddr_obj(uni->Address.lpSockaddr, uni->Address.iSockaddrLength));
-	}
-	for (any = adapters->FirstAnycastAddress; any; any = any->Next) {
+                rb_ary_push(list, sockaddr_obj(uni->Address.lpSockaddr, uni->Address.iSockaddrLength));
+        }
+        for (any = adapters->FirstAnycastAddress; any; any = any->Next) {
 #ifndef INET6
-	    if (any->Address.lpSockaddr->sa_family == AF_INET)
+            if (any->Address.lpSockaddr->sa_family == AF_INET)
 #else
-	    if (IS_IP_FAMILY(any->Address.lpSockaddr->sa_family))
+            if (IS_IP_FAMILY(any->Address.lpSockaddr->sa_family))
 #endif
-		rb_ary_push(list, sockaddr_obj(any->Address.lpSockaddr, any->Address.iSockaddrLength));
-	}
+                rb_ary_push(list, sockaddr_obj(any->Address.lpSockaddr, any->Address.iSockaddrLength));
+        }
     }
 
     FreeLibrary(h);
@@ -1941,6 +1856,8 @@ socket_s_ip_address_list(VALUE self)
 void
 Init_socket(void)
 {
+    rb_ext_ractor_safe(true);
+
     rsock_init_basicsocket();
 
     /*
@@ -2069,7 +1986,7 @@ Init_socket(void)
 
     /* for ext/socket/lib/socket.rb use only: */
     rb_define_private_method(rb_cSocket,
-			     "__connect_nonblock", sock_connect_nonblock, 2);
+                             "__connect_nonblock", sock_connect_nonblock, 2);
 
     rb_define_method(rb_cSocket, "bind", sock_bind, 1);
     rb_define_method(rb_cSocket, "listen", rsock_sock_listen, 1);
@@ -2077,7 +1994,7 @@ Init_socket(void)
 
     /* for ext/socket/lib/socket.rb use only: */
     rb_define_private_method(rb_cSocket,
-			     "__accept_nonblock", sock_accept_nonblock, 1);
+                             "__accept_nonblock", sock_accept_nonblock, 1);
 
     rb_define_method(rb_cSocket, "sysaccept", sock_sysaccept, 0);
 
@@ -2085,7 +2002,7 @@ Init_socket(void)
 
     /* for ext/socket/lib/socket.rb use only: */
     rb_define_private_method(rb_cSocket,
-			     "__recvfrom_nonblock", sock_recvfrom_nonblock, 4);
+                             "__recvfrom_nonblock", sock_recvfrom_nonblock, 4);
 
     rb_define_singleton_method(rb_cSocket, "socketpair", rsock_sock_s_socketpair, -1);
     rb_define_singleton_method(rb_cSocket, "pair", rsock_sock_s_socketpair, -1);
