@@ -3,30 +3,24 @@ require 'uri'
 require 'stringio'
 require 'time'
 
-module Kernel
-  private
-  alias open_uri_original_open open # :nodoc:
-  class << self
-    alias open_uri_original_open open # :nodoc:
-  end
-
+module URI
   # Allows the opening of various resources including URIs.
   #
   # If the first argument responds to the 'open' method, 'open' is called on
   # it with the rest of the arguments.
   #
-  # If the first argument is a string that begins with xxx://, it is parsed by
+  # If the first argument is a string that begins with <code>(protocol)://</code>, it is parsed by
   # URI.parse.  If the parsed object responds to the 'open' method,
   # 'open' is called on it with the rest of the arguments.
   #
-  # Otherwise, the original Kernel#open is called.
+  # Otherwise, Kernel#open is called.
   #
   # OpenURI::OpenRead#open provides URI::HTTP#open, URI::HTTPS#open and
   # URI::FTP#open, Kernel#open.
   #
   # We can accept URIs and strings that begin with http://, https:// and
   # ftp://. In these cases, the opened file object is extended by OpenURI::Meta.
-  def open(name, *rest, &block) # :doc:
+  def self.open(name, *rest, &block)
     if name.respond_to?(:open)
       name.open(*rest, &block)
     elsif name.respond_to?(:to_str) &&
@@ -34,10 +28,9 @@ module Kernel
           (uri = URI.parse(name)).respond_to?(:open)
       uri.open(*rest, &block)
     else
-      open_uri_original_open(name, *rest, &block)
+      super
     end
   end
-  module_function :open
 end
 
 # OpenURI is an easy-to-use wrapper for Net::HTTP, Net::HTTPS and Net::FTP.
@@ -46,14 +39,14 @@ end
 #
 # It is possible to open an http, https or ftp URL as though it were a file:
 #
-#   open("http://www.ruby-lang.org/") {|f|
+#   URI.open("http://www.ruby-lang.org/") {|f|
 #     f.each_line {|line| p line}
 #   }
 #
 # The opened file has several getter methods for its meta-information, as
 # follows, since it is extended by OpenURI::Meta.
 #
-#   open("http://www.ruby-lang.org/en") {|f|
+#   URI.open("http://www.ruby-lang.org/en") {|f|
 #     f.each_line {|line| p line}
 #     p f.base_uri         # <URI::HTTP:0x40e6ef2 URL:http://www.ruby-lang.org/en/>
 #     p f.content_type     # "text/html"
@@ -64,7 +57,7 @@ end
 #
 # Additional header fields can be specified by an optional hash argument.
 #
-#   open("http://www.ruby-lang.org/en/",
+#   URI.open("http://www.ruby-lang.org/en/",
 #     "User-Agent" => "Ruby/#{RUBY_VERSION}",
 #     "From" => "foo@bar.invalid",
 #     "Referer" => "http://www.ruby-lang.org/") {|f|
@@ -74,11 +67,11 @@ end
 # The environment variables such as http_proxy, https_proxy and ftp_proxy
 # are in effect by default. Here we disable proxy:
 #
-#   open("http://www.ruby-lang.org/en/", :proxy => nil) {|f|
+#   URI.open("http://www.ruby-lang.org/en/", :proxy => nil) {|f|
 #     # ...
 #   }
 #
-# See OpenURI::OpenRead.open and Kernel#open for more on available options.
+# See OpenURI::OpenRead.open and URI.open for more on available options.
 #
 # URI objects can be opened in a similar way.
 #
@@ -108,6 +101,7 @@ module OpenURI
     :ssl_verify_mode => nil,
     :ftp_active_mode => false,
     :redirect => true,
+    :encoding => nil,
   }
 
   def OpenURI.check_options(options) # :nodoc:
@@ -141,6 +135,12 @@ module OpenURI
       encoding, = $1,Encoding.find($1) if $1
       mode = nil
     end
+    if options.has_key? :encoding
+      if !encoding.nil?
+        raise ArgumentError, "encoding specified twice"
+      end
+      encoding = Encoding.find(options[:encoding])
+    end
 
     unless mode == nil ||
            mode == 'r' || mode == 'rb' ||
@@ -340,6 +340,7 @@ module OpenURI
           if options[:progress_proc] && Net::HTTPSuccess === resp
             options[:progress_proc].call(buf.size)
           end
+          str.clear
         }
       }
     }
@@ -528,17 +529,16 @@ module OpenURI
     # It can be used to guess charset.
     #
     # If charset parameter and block is not given,
-    # nil is returned except text type in HTTP.
-    # In that case, "iso-8859-1" is returned as defined by RFC2616 3.7.1.
+    # nil is returned except text type.
+    # In that case, "utf-8" is returned as defined by RFC6838 4.2.1
     def charset
       type, *parameters = content_type_parse
       if pair = parameters.assoc('charset')
         pair.last.downcase
       elsif block_given?
         yield
-      elsif type && %r{\Atext/} =~ type &&
-            @base_uri && /\Ahttp\z/i =~ @base_uri.scheme
-        "iso-8859-1" # RFC2616 3.7.1
+      elsif type && %r{\Atext/} =~ type
+        "utf-8" # RFC6838 4.2.1
       else
         nil
       end
@@ -638,8 +638,8 @@ module OpenURI
     #  is called before actual transfer is started.
     #  It takes one argument, which is expected content length in bytes.
     #
-    #  If two or more transfer is done by HTTP redirection, the procedure
-    #  is called only one for a last transfer.
+    #  If two or more transfers are performed by HTTP redirection, the
+    #  procedure is called only once for the last transfer.
     #
     #  When expected content length is unknown, the procedure is called with
     #  nil.  This happens when the HTTP response has no Content-Length header.
@@ -721,7 +721,7 @@ module OpenURI
       OpenURI.open_uri(self, *rest, &block)
     end
 
-    # OpenURI::OpenRead#read([options]) reads a content referenced by self and
+    # OpenURI::OpenRead#read([ options ]) reads a content referenced by self and
     # returns the content as string.
     # The string is extended with OpenURI::Meta.
     # The argument +options+ is same as OpenURI::OpenRead#open.
@@ -750,7 +750,12 @@ module URI
         OpenURI.open_http(buf, self, proxy, options)
         return
       end
-      require 'net/ftp'
+
+      begin
+        require 'net/ftp'
+      rescue LoadError
+        abort "net/ftp is not found. You may need to `gem install net-ftp` to install net/ftp."
+      end
 
       path = self.path
       path = path.sub(%r{\A/}, '%2F') # re-encode the beginning slash because uri library decodes it.
