@@ -17,8 +17,8 @@ class TestRDocOptions < RDoc::TestCase
   end
 
   def test_check_files
-    skip "assumes UNIX permission model" if /mswin|mingw/ =~ RUBY_PLATFORM
-    skip "assumes that euid is not root" if Process.euid == 0
+    omit "assumes UNIX permission model" if /mswin|mingw/ =~ RUBY_PLATFORM
+    omit "assumes that euid is not root" if Process.euid == 0
 
     out, err = capture_output do
       temp_dir do
@@ -55,11 +55,10 @@ class TestRDocOptions < RDoc::TestCase
     refute @options.dry_run
   end
 
-  def test_encode_with
-    coder = {}
-    class << coder; alias add []=; end
+  def test_to_yaml
+    RDoc.load_yaml
 
-    @options.encode_with coder
+    coder = YAML.load(@options.to_yaml)
 
     encoding = 'UTF-8'
 
@@ -89,10 +88,11 @@ class TestRDocOptions < RDoc::TestCase
     assert_equal expected, coder
   end
 
-  def test_encode_with_trim_paths
+  def test_to_yaml_trim_paths
+    RDoc.load_yaml
+
     subdir = nil
-    coder = {}
-    class << coder; alias add []=; end
+    coder = nil
 
     temp_dir do |dir|
       FileUtils.mkdir 'project'
@@ -113,7 +113,7 @@ class TestRDocOptions < RDoc::TestCase
           --include /
         ]
 
-        @options.encode_with coder
+        coder = YAML.load(@options.to_yaml)
       end
     end
 
@@ -145,7 +145,9 @@ class TestRDocOptions < RDoc::TestCase
 
     @options.encoding = Encoding::IBM437
 
-    options = YAML.load YAML.dump @options
+    options = @options.to_yaml
+    options = YAML.safe_load(options, permitted_classes: [Symbol])
+    options = RDoc::Options.new(options)
 
     assert_equal Encoding::IBM437, options.encoding
   end
@@ -154,14 +156,15 @@ class TestRDocOptions < RDoc::TestCase
     RDoc.load_yaml
 
     yaml = <<-YAML
---- !ruby/object:RDoc::Options
+---
 static_path:
 - /etc
 rdoc_include:
 - /etc
     YAML
 
-    options = YAML.load yaml
+    options = YAML.safe_load(yaml, permitted_classes: [Symbol])
+    options = RDoc::Options.new(options)
 
     assert_empty options.rdoc_include
     assert_empty options.static_path
@@ -243,6 +246,7 @@ rdoc_include:
 
   def test_parse_default
     @options.parse []
+    @options.finish
 
     assert_equal RDoc::Generator::Darkfish,             @options.generator
     assert_equal 'darkfish',                            @options.template
@@ -502,13 +506,20 @@ rdoc_include:
 
     out, err = capture_output do
       @options.parse %W[--page-dir #{Dir.tmpdir}]
+      @options.finish
     end
 
     assert_empty out
     assert_empty err
 
-    expected =
-      Pathname(Dir.tmpdir).expand_path.relative_path_from @options.root
+    expected = nil
+    begin
+      expected =
+        Pathname(Dir.tmpdir).expand_path.relative_path_from @options.root
+    rescue ArgumentError
+      # On Windows, sometimes crosses different drive letters.
+      expected = Pathname(Dir.tmpdir).expand_path
+    end
 
     assert_equal expected,     @options.page_dir
     assert_equal [Dir.tmpdir], @options.files
@@ -524,6 +535,7 @@ rdoc_include:
 
       out, err = capture_output do
         @options.parse %W[--page-dir #{abs_page_dir} --root #{abs_root}]
+        @options.finish
       end
 
       assert_empty out
@@ -552,6 +564,8 @@ rdoc_include:
     assert_empty err
 
     assert_equal Pathname(Dir.tmpdir), @options.root
+
+    @options.finish
     assert_includes @options.rdoc_include, @options.root.to_s
   end
 
@@ -596,6 +610,7 @@ rdoc_include:
     assert_empty out
     assert_equal "could not find template NONEXISTENT\n", err
 
+    @options.finish
     assert_equal 'darkfish', @options.template
     assert_match %r%rdoc/generator/template/darkfish$%, @options.template_dir
   end
@@ -626,6 +641,21 @@ rdoc_include:
     $LOAD_PATH.replace orig_LOAD_PATH
   end
 
+  def test_parse_template_stylesheets
+    css = nil
+    Dir.mktmpdir do |dir|
+      css = File.join(dir, "hoge.css")
+      File.write(css, "")
+      out, err = capture_output do
+        @options.parse %W[--template-stylesheets #{css}]
+      end
+
+      assert_empty out
+      assert_empty err
+    end
+    assert_include @options.template_stylesheets, css
+  end
+
   def test_parse_visibility
     @options.parse %w[--visibility=public]
     assert_equal :public, @options.visibility
@@ -647,6 +677,7 @@ rdoc_include:
     Dir.chdir tmpdir do
       e = assert_raise SystemExit do
         @options.parse %w[--write-options]
+        @options.finish
       end
 
       assert_equal 0, e.status
@@ -743,7 +774,9 @@ rdoc_include:
 
       assert File.exist? '.rdoc_options'
 
-      assert_equal @options, YAML.load(File.read('.rdoc_options'))
+      options = File.read('.rdoc_options')
+      options = YAML.safe_load(options, permitted_classes: [Symbol])
+      assert_equal @options, RDoc::Options.new(options)
     end
   end
 
@@ -771,4 +804,76 @@ rdoc_include:
     @options.visibility = :all
     assert_equal :private, @options.visibility
   end
+
+  def test_load_options
+    temp_dir do
+      options = RDoc::Options.new
+      options.markup = 'tomdoc'
+      options.write_options
+
+      options = RDoc::Options.load_options
+
+      assert_equal 'tomdoc', options.markup
+    end
+  end
+
+  def test_load_options_invalid
+    temp_dir do
+      File.open '.rdoc_options', 'w' do |io|
+        io.write "a: !ruby.yaml.org,2002:str |\nfoo"
+      end
+
+      e = assert_raise RDoc::Error do
+        RDoc::Options.load_options
+      end
+
+      options_file = File.expand_path '.rdoc_options'
+      assert_equal "#{options_file} is not a valid rdoc options file", e.message
+    end
+  end
+
+  def test_load_options_empty_file
+    temp_dir do
+      File.open '.rdoc_options', 'w' do |io|
+      end
+
+      options = RDoc::Options.load_options
+
+      assert_equal 'rdoc', options.markup
+    end
+  end
+
+  def test_load_options_partial_override
+    temp_dir do
+      File.open '.rdoc_options', 'w' do |io|
+        io.puts "markup: Markdown"
+        io.puts "encoding: iso-8859-1"
+        io.puts "static_path: [static]"
+        io.puts "rdoc_include: [.]"
+        io.puts "page_dir: pages"
+      end
+
+      options = RDoc::Options.load_options
+
+      assert_equal 'Markdown', options.markup
+      assert_equal Encoding::ISO_8859_1, options.encoding
+      assert_equal ["static"], options.static_path
+      assert_equal ["."], options.rdoc_include
+      assert_equal "pages", options.page_dir
+    end
+  end
+
+  def test_load_options_no_file
+    temp_dir do
+      options = RDoc::Options.load_options
+
+      assert_kind_of RDoc::Options, options
+    end
+  end
+
+  class DummyCoder < Hash
+    alias add :[]=
+    def tag=(tag)
+    end
+  end
 end
